# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

# Docker image for running Cebes locally
LOCAL_HTTP_SERVER_REPO = 'cebesio/http-server'
LOCAL_HTTP_SERVER_TAG = '0.10.0'

LOCAL_REPOSITORY_REPO = 'cebesio/pipeline-repo'
LOCAL_REPOSITORY_TAG = '0.10.0'
